#include "paths/paths.h"
#include "butterflies_populations.h"
#include "modules/module_header.h"
#include "actions/cats_actions.h"

#include "butterflies_overlays.h"
#include "butterflies_main.h"
#include "inline_overlays.h"
#include "inline_population.h"
#include "inline_butterflies.h"


bool butterflies_habitat_ok(const struct cats_configuration *conf, cats_dt_coord row, cats_dt_coord col)
{
        // we have no habitat layer -> all cells have suitable habitat
        if (conf->overlays.overlay[OL_HABITAT_TYPE_CC].enabled == false) {
                return true;
        }
        // habitat layer exists and the carrying capacity is > 0 -> this cell's habitat is suitable
        if (get_overlay_habitat_cc_value(conf, row, col) > 0) {
                return true;
        }

        return false;
}


bool butterflies_resources_ok(const struct cats_configuration *conf, cats_dt_coord row, cats_dt_coord col)
{
        // we have no resource layer -> all cells have the resource available
        if (conf->overlays.overlay[OL_RESOURCE].enabled == false) {
                return true;
        }

        // resource layer exists and the resource is > 0 -> this cell's resource is available
        if (get_overlay_resource_value(conf, row, col) > 0) {
                return true;
        }

        return false;
}


enum action_status bf_grid_overlay_update(const struct cats_configuration *conf, struct cats_grid *grid)
{
        if (conf->overlays.have_overlays == false) {
                return ACTION_NOT_RUN;
        }

        int module_id = CATS_MODULE_ID;
        struct grid_data_butterflies *data = grid->grid_modules[module_id].module_data;


        const cats_dt_coord rows = grid->dimension.rows;
        const cats_dt_coord cols = grid->dimension.cols;
        int64_t cells_habitat_ok = 0;
        int64_t cells_excluded = 0;
        int64_t cells_resource_ok = 0;
        int64_t cells_habitat_and_resource_ok = 0;
        int64_t cells_eggs_removed = 0;
        int64_t cells_adults_removed = 0;

        for (cats_dt_coord row = 0; row < rows; row++) {
                for (cats_dt_coord col = 0; col < cols; col++) {
                        int32_t cell_info = BF_CELL_CLEAR;

                        float eggs = butterflies_get_eggs(data, row, col);
                        if (cell_excluded_by_overlay(conf, row, col)) {
                                cell_info |= BF_CELL_EXCLUDED;
                                butterflies_set_info_layer(data, row, col, cell_info);
                                if (eggs) cells_eggs_removed += 1;
                                if (get_adult_population(grid, row, col)) cells_adults_removed += 1;
                                butterflies_set_eggs(data, row, col, 0.0f);
                                set_population_ignore_cc(grid, row, col, 0);
                                cells_excluded += 1;
                                continue;
                        }

                        if (butterflies_habitat_ok(conf, row, col)) {
                                cell_info |= BF_CELL_HABITAT_OK;
                                cells_habitat_ok += 1;
                        }

                        if (butterflies_resources_ok(conf, row, col)) {
                                cell_info |= BF_CELL_RESOURCE_AVAILABLE;
                                cells_resource_ok += 1;
                        }


                        if ((cell_info & BF_CELL_HABITAT_OK) && (cell_info & BF_CELL_RESOURCE_AVAILABLE)) {
                                cell_info |= BF_CELL_VALID_DISPERSAL_TARGET;
                                cells_habitat_and_resource_ok += 1;

                        } else {
                                if (eggs) cells_eggs_removed += 1;
                                if (get_adult_population(grid, row, col)) cells_adults_removed += 1;
                                butterflies_set_eggs(data, row, col, 0.0f);
                                set_population_ignore_cc(grid, row, col, 0);
                        }

                        butterflies_set_info_layer(data, row, col, cell_info);
                }
        }

        int64_t cells_with_eggs = 0;
        double total_eggs = 0;
        for (cats_dt_coord row = 0; row < rows; row++) {
                for (cats_dt_coord col = 0; col < cols; col++) {
                        float eggs = butterflies_get_eggs(data, row, col);
                        if (eggs > 0) {
                                cells_with_eggs += 1;
                                total_eggs += eggs;
                        }

                }
        }

        log_message(LOG_INFO,
                    "Overlay update: %ld excluded, %ld habitat ok, %ld resource ok, %ld habitat + resource ok",
                    cells_excluded, cells_habitat_ok, cells_resource_ok, cells_habitat_and_resource_ok);
        log_message(LOG_INFO, "Overlay update: %ld cells with eggs, total %f eggs, average %f eggs per cell",
                    cells_with_eggs, total_eggs, total_eggs / (float) cells_with_eggs);
        log_message(LOG_INFO, "Overlay update: %ld cells with eggs removed", cells_eggs_removed);
        log_message(LOG_INFO, "Overlay update: %ld cells with adults removed", cells_adults_removed);


        return ACTION_RUN;
}

